@@ -25,9 +25,13 @@ type Dialog interface {
 	SendMessage(message.HBCIMessage) (message.BankMessage, error)
 }
 
-const initialDialogID = "0"
-const initialClientSystemID = "0"
-const anonymousClientID = "9999999999"
+const (
+	initialDialogID                 = "0"
+	initialClientSystemID           = "0"
+	initialBankParameterDataVersion = 0
+	initialUserParameterDataVersion = 0
+	anonymousClientID               = "9999999999"
+)
 
 func newDialog(
 	bankID domain.BankID,
@@ -67,7 +71,7 @@ type dialog struct {
 	securityFn        string
 	signatureProvider message.SignatureProvider
 	cryptoProvider    message.CryptoProvider
-	BankParameterData domain.BankParameterData
+	BankParameterData BankParameterData
 	hbciVersion       segment.HBCIVersion
 	supportedSegments []segment.VersionedSegment
 }
@@ -126,15 +130,31 @@ func (d *dialog) SendMessage(clientMessage message.HBCIMessage) (message.BankMes
 		}
 	}
 	if len(errors) > 0 {
-		return nil, fmt.Errorf("Institute returned errors:\n%s", strings.Join(errors, "\n"))
+		return nil, fmt.Errorf("institute returned errors:\n%s", strings.Join(errors, "\n"))
 	}
 	return decryptedMessage, nil
 }
 
+func (d *dialog) SyncUserParameterData() error {
+	internal.Info.Printf("Initializing dialog")
+	err := d.init()
+	if err != nil {
+		return err
+	}
+	defer func() {
+		internal.Info.Printf("Ending dialog")
+		logErr(d.end())
+	}()
+	return nil
+}
+
 func (d *dialog) SyncClientSystemID() (string, error) {
 	syncMessage := message.NewSynchronisationMessage(d.hbciVersion)
 	syncMessage.Identification = segment.NewIdentificationSegment(d.BankID, d.clientID, initialClientSystemID, true)
-	syncMessage.ProcessingPreparation = segment.NewProcessingPreparationSegment(0, 0, 1)
+	syncMessage.ProcessingPreparation = segment.NewProcessingPreparationSegmentV3(
+		initialBankParameterDataVersion, initialUserParameterDataVersion, domain.German,
+	)
+	syncMessage.TanRequest = d.hbciVersion.TanProcess4Request(segment.IdentificationID)
 	syncMessage.Sync = d.hbciVersion.SynchronisationRequest(segment.SyncModeAquireClientID)
 	syncMessage.BasicMessage = d.newBasicMessage(syncMessage)
 	signedSyncMessage, err := syncMessage.Sign(d.signatureProvider)
@@ -149,19 +169,20 @@ func (d *dialog) SyncClientSystemID() (string, error) {
 
 	decryptedMessage, err := d.request(encryptedSyncMessage)
 	if err != nil {
-		return "", fmt.Errorf("Error while extracting encrypted message: %v", err)
+		return "", fmt.Errorf("error while extracting encrypted message: %v", err)
 	}
 
 	messageHeader := decryptedMessage.MessageHeader()
 	if messageHeader == nil {
-		return "", fmt.Errorf("Malformed response message: %q", decryptedMessage)
+		return "", fmt.Errorf("malformed response message: %q", decryptedMessage)
 	}
 	d.dialogID = messageHeader.DialogID.Val()
-	d.supportedSegments = decryptedMessage.SupportedSegments()
-
 	var errors []string
 	acknowledgements := decryptedMessage.Acknowledgements()
 	for _, ack := range acknowledgements {
+		if ack.IsSuccess() {
+			internal.Debug.Printf("%v\n", ack)
+		}
 		if ack.IsWarning() {
 			internal.Info.Printf("%v\n", ack)
 		}
@@ -170,7 +191,10 @@ func (d *dialog) SyncClientSystemID() (string, error) {
 		}
 	}
 	if len(errors) > 0 {
-		return "", fmt.Errorf("Institute returned errors:\n%s", strings.Join(errors, "\n"))
+		return "", fmt.Errorf("institute returned errors:\n%s", strings.Join(errors, "\n"))
+	}
+	if err := d.updateSecurityFunctionIfNeeded(decryptedMessage); err != nil {
+		return "", fmt.Errorf("error updating security function: %w", err)
 	}
 
 	syncResponse := decryptedMessage.FindSegment("HISYN")
@@ -178,7 +202,7 @@ func (d *dialog) SyncClientSystemID() (string, error) {
 		syncSegment := syncResponse.(segment.SynchronisationResponse)
 		d.SetClientSystemID(syncSegment.ClientSystemID())
 	} else {
-		return "", fmt.Errorf("Malformed message: missing unmarshaler for SynchronisationResponse")
+		return "", fmt.Errorf("malformed message: missing unmarshaler for SynchronisationResponse")
 	}
 
 	err = d.parseBankParameterData(decryptedMessage)
@@ -207,7 +231,7 @@ func (d *dialog) SendAnonymousMessage(clientMessage message.HBCIMessage) (messag
 	defer func() { logErr(d.anonymousEnd()) }()
 	// TODO: add checks if job needs signature or not
 	requestMessage := d.newBasicMessage(clientMessage)
-	requestMessage.SetNumbers()
+	requestMessage.SetSegmentPositions()
 	bankMessage, err := d.request(requestMessage)
 	if err != nil {
 		return nil, err
@@ -233,16 +257,18 @@ func (d *dialog) anonymousInit() error {
 	d.messageCount = 0
 	initMessage := message.NewDialogInitializationClientMessage(d.hbciVersion)
 	initMessage.Identification = segment.NewIdentificationSegment(d.BankID, anonymousClientID, initialClientSystemID, false)
-	initMessage.ProcessingPreparation = segment.NewProcessingPreparationSegment(d.BankParameterDataVersion(), d.UserParameterDataVersion(), d.Language)
+	initMessage.ProcessingPreparation = segment.NewProcessingPreparationSegmentV3(
+		d.BankParameterDataVersion(), d.UserParameterDataVersion(), d.Language,
+	)
 	initMessage.BasicMessage = d.newBasicMessage(initMessage)
-	initMessage.SetNumbers()
+	initMessage.SetSegmentPositions()
 	bankMessage, err := d.request(initMessage)
 	if err != nil {
 		return err
 	}
 	messageHeader := bankMessage.MessageHeader()
 	if messageHeader == nil {
-		return fmt.Errorf("Malformed response message: %q", bankMessage)
+		return fmt.Errorf("malformed response message: %q", bankMessage)
 	}
 	d.dialogID = messageHeader.DialogID.Val()
 
@@ -256,7 +282,7 @@ func (d *dialog) anonymousInit() error {
 		return err
 	}
 
-	bankInfoMessage := bankMessage.FindSegment("HIKIM")
+	bankInfoMessage := bankMessage.FindSegment(segment.BankAnnouncementID)
 	if bankInfoMessage != nil {
 		bankInfoSegment := bankInfoMessage.(*segment.BankAnnouncementSegment)
 		internal.Info.Printf("INFO:\n%s\n%s\n", bankInfoSegment.Subject.Val(), bankInfoSegment.Body.Val())
@@ -281,7 +307,7 @@ func (d *dialog) anonymousInit() error {
 func (d *dialog) anonymousEnd() error {
 	dialogEnd := message.NewDialogFinishingMessage(d.hbciVersion, d.dialogID)
 	dialogEnd.BasicMessage = d.newBasicMessage(dialogEnd)
-	dialogEnd.SetNumbers()
+	dialogEnd.SetSegmentPositions()
 
 	decryptedMessage, err := d.request(dialogEnd)
 	if err != nil {
@@ -314,7 +340,10 @@ func (d *dialog) init() error {
 	d.messageCount = 0
 	initMessage := message.NewDialogInitializationClientMessage(d.hbciVersion)
 	initMessage.Identification = segment.NewIdentificationSegment(d.BankID, d.clientID, d.ClientSystemID, true)
-	initMessage.ProcessingPreparation = segment.NewProcessingPreparationSegment(d.BankParameterDataVersion(), d.UserParameterDataVersion(), d.Language)
+	initMessage.ProcessingPreparation = segment.NewProcessingPreparationSegmentV3(
+		d.BankParameterDataVersion(), d.UserParameterDataVersion(), d.Language,
+	)
+	initMessage.TanRequest = d.hbciVersion.TanProcess4Request(segment.IdentificationID)
 	initMessage.BasicMessage = d.newBasicMessage(initMessage)
 	signedInitMessage, err := initMessage.Sign(d.signatureProvider)
 	if err != nil {
@@ -327,11 +356,11 @@ func (d *dialog) init() error {
 
 	decryptedMessage, err := d.request(encryptedInitMessage)
 	if err != nil {
-		return fmt.Errorf("Error while initializing dialog: %v", err)
+		return fmt.Errorf("error while initializing dialog: %v", err)
 	}
 	messageHeader := decryptedMessage.MessageHeader()
 	if messageHeader == nil {
-		return fmt.Errorf("Malformed response message: %q", decryptedMessage)
+		return fmt.Errorf("malformed response message: %q", decryptedMessage)
 	}
 	d.dialogID = messageHeader.DialogID.Val()
 
@@ -345,23 +374,20 @@ func (d *dialog) init() error {
 		return err
 	}
 
-	bankInfoMessage := decryptedMessage.FindSegment("HIKIM")
+	bankInfoMessage := decryptedMessage.FindSegment(segment.BankAnnouncementID)
 	if bankInfoMessage != nil {
 		bankInfoSegment := bankInfoMessage.(*segment.BankAnnouncementSegment)
 		internal.Info.Printf("INFO:\n%s\n%s\n", bankInfoSegment.Subject.Val(), bankInfoSegment.Body.Val())
 	}
 
-	newSecurityFn := d.securityFn
-	errors := make([]string, 0)
+	var errors []string
 	acknowledgements := decryptedMessage.Acknowledgements()
 	for _, ack := range acknowledgements {
-		if ack.Code == element.AcknowledgementSupportedSecurityFunction {
-			supportedSecurityFns := ack.Params
-			if len(supportedSecurityFns) != 0 {
-				internal.Info.Printf("Supported securityFunctions: %q\n", supportedSecurityFns)
-				// TODO: proper handling of each case, see FINTS3.0 docu
-				newSecurityFn = supportedSecurityFns[0]
-			}
+		if ack.IsSuccess() {
+			internal.Debug.Printf("%v\n", ack)
+		}
+		if ack.IsWarning() {
+			internal.Info.Printf("%v\n", ack)
 		}
 		if ack.IsError() {
 			errors = append(errors, ack.String())
@@ -370,16 +396,9 @@ func (d *dialog) init() error {
 	if len(errors) > 0 {
 		return fmt.Errorf("DialogEnd: Institute returned errors:\n%s", strings.Join(errors, "\n"))
 	}
-	if d.securityFn != newSecurityFn {
-		err = d.end()
-		if err != nil {
-			return err
-		}
-		d.SetSecurityFunction(newSecurityFn)
-		err = d.init()
-		if err != nil {
-			return err
-		}
+
+	if err := d.updateSecurityFunctionIfNeeded(decryptedMessage); err != nil {
+		return fmt.Errorf("error updating security function: %w", err)
 	}
 
 	return nil
@@ -428,13 +447,95 @@ func (d *dialog) newBasicMessage(hbciMessage message.HBCIMessage) *message.Basic
 	return clientMessage
 }
 
-func (d *dialog) parseBankParameterData(bankMessage message.BankMessage) error {
-	bankParamData := bankMessage.FindSegment("HIBPA")
-	if bankParamData != nil {
-		paramSegment := bankParamData.(segment.CommonBankParameter)
-		d.BankParameterData = paramSegment.BankParameterData()
+func (d *dialog) updateSecurityFunctionIfNeeded(message message.BankMessage) error {
+	if !d.hasSupportedSecurityAcknowledgement(message) {
+		return nil
 	}
-	pinTanTransactions := bankMessage.FindSegment("DIPINS")
+	supportedSecurityFns, ok := d.supportedSecurityFunctionsFromBankMessage(message)
+	if !ok {
+		return fmt.Errorf("no supported security function implemented")
+	}
+	oldSecurityFn := d.securityFn
+	newSecurityFn := ""
+	newSecurityFnName := ""
+	// TODO: proper handling of each case, see FINTS3.0 docu -> Ask user which to use
+	for fn, name := range supportedSecurityFns {
+		newSecurityFn = fn
+		newSecurityFnName = name
+		break
+	}
+	if oldSecurityFn != newSecurityFn {
+		internal.Info.Printf(
+			"New supported security function found. Setting new security function %q (%s).", newSecurityFnName, newSecurityFn,
+		)
+		d.SetSecurityFunction(newSecurityFn)
+
+	}
+	return nil
+}
+
+func (d *dialog) hasSupportedSecurityAcknowledgement(message message.BankMessage) bool {
+	rawTanBPDSegment := message.FindSegment(segment.TanBankParameterID)
+	if rawTanBPDSegment == nil {
+		return false
+	}
+	for _, ack := range message.Acknowledgements() {
+		if ack.Code == element.AcknowledgementSupportedSecurityFunction {
+			return true
+		}
+	}
+	return false
+}
+
+func (d *dialog) supportedSecurityFunctionsFromBankMessage(message message.BankMessage) (map[string]string, bool) {
+	acknowledgements := message.Acknowledgements()
+	var supportedSecurityFns []string
+	for _, ack := range acknowledgements {
+		if ack.Code == element.AcknowledgementSupportedSecurityFunction {
+			supportedSecurityFns = ack.Params
+			break
+		}
+	}
+	if len(supportedSecurityFns) == 0 {
+		return nil, false
+	}
+	securityFunctions := map[string]string{}
+	rawSegment := message.FindSegment(segment.TanBankParameterID)
+	if rawSegment == nil {
+		return nil, false
+	}
+	availableSecurityFns := map[string]*element.Tan2StepSubmissionProcessParameterV6{}
+	for _, pp := range rawSegment.(*segment.TanBankParameterV6).Tan2StepSubmissionParameter.ProcessParameters.GroupDataElements() {
+		tan2StepPP := pp.(*element.Tan2StepSubmissionProcessParameterV6)
+		fn := tan2StepPP.SecurityFunction.Val()
+		availableSecurityFns[fn] = tan2StepPP
+	}
+	for _, sf := range supportedSecurityFns {
+		if pp, ok := availableSecurityFns[sf]; ok {
+			securityFunctions[sf] = pp.TwoStepProcessName.Val()
+		}
+	}
+	if len(securityFunctions) == 0 {
+		return nil, false
+	}
+	return securityFunctions, true
+}
+
+func (d *dialog) parseBankParameterData(bankMessage message.BankMessage) error {
+	bankParamData := bankMessage.FindSegment(segment.CommonBankParameterID)
+	if bankParamData == nil {
+		return nil
+	}
+	paramSegment, ok := bankParamData.(segment.CommonBankParameter)
+	if !ok {
+		return fmt.Errorf("error converting common bank parameter data")
+	}
+	d.supportedSegments = bankMessage.SupportedSegments()
+	d.BankParameterData = BankParameterData{
+		BankParameterData:          paramSegment.BankParameterData(),
+		SupportedSegmentParameters: make([]SegmentParameter, len(d.supportedSegments)),
+	}
+	pinTanTransactions := bankMessage.FindSegment(segment.PinTanBankParameterID)
 	if pinTanTransactions != nil {
 		pinTanTransactionSegment := pinTanTransactions.(segment.PinTanBusinessTransactionParams)
 		pinTransactions := make(map[string]bool)
@@ -443,25 +544,32 @@ func (d *dialog) parseBankParameterData(bankMessage message.BankMessage) error {
 		}
 		d.BankParameterData.PinTanBusinessTransactions = pinTransactions
 	}
+	for i, s := range d.supportedSegments {
+		param := SegmentParameter{
+			VersionedSegment: s,
+		}
+		parameterData := bankMessage.FindSegment(s.ID)
+		if parameterData != nil && parameterData.Header().Version.Val() == s.Version {
+			param.Parameters = parameterData
+		}
+		d.BankParameterData.SupportedSegmentParameters[i] = param
+	}
 	return nil
 }
 
 func (d *dialog) parseUserParameterData(bankMessage message.BankMessage) error {
-	userParamData := bankMessage.FindSegment("HIUPA")
+	userParamData := bankMessage.FindSegment(segment.CommonUserParameterDataID)
 	if userParamData != nil {
 		paramSegment := userParamData.(segment.CommonUserParameterData)
 		d.UserParameterData = paramSegment.UserParameterData()
 		d.clientID = d.UserParameterData.UserID
 	}
 
-	accountData := bankMessage.FindSegments("HIUPD")
-	if accountData != nil {
-		for _, acc := range accountData {
-			infoSegment := acc.(segment.AccountInformation)
-			d.Accounts = append(d.Accounts, infoSegment.Account())
-		}
+	accountData := bankMessage.FindSegments(segment.AccountInformationID)
+	for _, acc := range accountData {
+		infoSegment := acc.(segment.AccountInformation)
+		d.Accounts = append(d.Accounts, infoSegment.Account())
 	}
-
 	return nil
 }
 
@@ -484,11 +592,11 @@ func (d *dialog) request(clientMessage message.ClientMessage) (message.BankMessa
 
 	response, err := d.transport.Do(request)
 	if err != nil {
-		return nil, fmt.Errorf("Transport#Do: %v", err)
+		return nil, fmt.Errorf("error executing Transport request: %v", err)
 	}
 	response, err = transport.ReadResponse(bufio.NewReader(response.Body), response.Request)
 	if err != nil {
-		return nil, fmt.Errorf("Transport#ReadRequest: %v", err)
+		return nil, fmt.Errorf("error reading response from Transport: %v", err)
 	}
 
 	var bankMessage message.BankMessage
@@ -500,7 +608,7 @@ func (d *dialog) request(clientMessage message.ClientMessage) (message.BankMessa
 
 		decryptedMessage, err := encMessage.Decrypt(d.cryptoProvider)
 		if err != nil {
-			return nil, fmt.Errorf("Error while decrypting message: %v", err)
+			return nil, fmt.Errorf("error while decrypting message: %v", err)
 		}
 		internal.Debug.Printf("Response:\n %s\n", decryptedMessage.MessageHeader())
 		bankMessage = decryptedMessage
@@ -517,14 +625,14 @@ func (d *dialog) request(clientMessage message.ClientMessage) (message.BankMessa
 }
 
 func (d *dialog) extractEncryptedMessage(response *transport.Response) (*message.EncryptedMessage, error) {
-	messageHeader := response.FindSegment("HNHBK")
+	messageHeader := response.FindSegment(segment.MessageHeaderID)
 	if messageHeader == nil {
-		return nil, fmt.Errorf("Malformed response: missing Message Header")
+		return nil, fmt.Errorf("malformed response: missing Message Header")
 	}
 	header := &segment.MessageHeaderSegment{}
 	err := header.UnmarshalHBCI(messageHeader)
 	if err != nil {
-		return nil, fmt.Errorf("Error while unmarshaling message header: %v", err)
+		return nil, fmt.Errorf("error while unmarshaling message header: %v", err)
 	}
 	// TODO: parse messageEnd
 	// TODO: parse encryptionHeader
@@ -536,19 +644,19 @@ func (d *dialog) extractEncryptedMessage(response *transport.Response) (*message
 		encSegment := &segment.EncryptedDataSegment{}
 		err = encSegment.UnmarshalHBCI(encryptedData)
 		if err != nil {
-			return nil, fmt.Errorf("Error while unmarshaling encrypted data: %v", err)
+			return nil, fmt.Errorf("error while unmarshaling encrypted data: %v", err)
 		}
 		encMessage.EncryptedData = encSegment
 	} else {
-		return nil, fmt.Errorf("Malformed response: missing encrypted data: \n%v", response)
+		return nil, fmt.Errorf("malformed response: missing encrypted data: \n%v", response)
 	}
 	return encMessage, nil
 }
 
 func extractUnencryptedMessage(response *transport.Response) (message.BankMessage, error) {
-	messageHeader := response.FindSegment("HNHBK")
+	messageHeader := response.FindSegment(segment.MessageHeaderID)
 	if messageHeader == nil {
-		return nil, fmt.Errorf("Malformed response: missing Message Header")
+		return nil, fmt.Errorf("malformed response: missing Message Header")
 	}
 	header := &segment.MessageHeaderSegment{}
 	err := header.UnmarshalHBCI(messageHeader)
@@ -558,7 +666,7 @@ func extractUnencryptedMessage(response *transport.Response) (message.BankMessag
 	// TODO: parse messageEnd
 	decryptedMessage, err := message.NewDecryptedMessage(header, nil, response.MarshaledResponse)
 	if err != nil {
-		return nil, fmt.Errorf("Error while unmarshaling unencrypted data: %v", err)
+		return nil, fmt.Errorf("error while unmarshaling unencrypted data: %v", err)
 	}
 	return decryptedMessage, nil
 }
