@@ -4,6 +4,8 @@ import "github.com/mitch000001/go-hbci/element"
 
 //go:generate go run ../cmd/unmarshaler/unmarshaler_generator.go -segment BankAnnouncementSegment
 
+const BankAnnouncementID = "HIKIM"
+
 type BankAnnouncementSegment struct {
 	Segment
 	Subject *element.AlphaNumericDataElement
@@ -11,7 +13,7 @@ type BankAnnouncementSegment struct {
 }
 
 func (b *BankAnnouncementSegment) Version() int         { return 2 }
-func (b *BankAnnouncementSegment) ID() string           { return "HIKIM" }
+func (b *BankAnnouncementSegment) ID() string           { return BankAnnouncementID }
 func (b *BankAnnouncementSegment) referencedId() string { return "" }
 func (b *BankAnnouncementSegment) sender() string       { return senderBank }
 
